#!/usr/bin/env python3

# Copyright (C) 2015-9 Neil Freeman

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Census Geocoder wrapper
see http://geocoding.geo.census.gov/geocoder/Geocoding_Services_API.pdf
Accepts either named `lat` and `lng` or x and y inputs.
"""
import csv
import io

import requests
from requests.exceptions import RequestException
from requests_toolbelt.multipart.encoder import MultipartEncoder


DEFAULT_BENCHMARK = "Public_AR_Current"
DEFAULT_VINTAGE = "Current_Current"


class CensusGeocode:
    """Fetch results from the Census Geocoder"""

    _url = "https://geocoding.geo.census.gov/geocoder/{returntype}/{searchtype}"
    returntypes = ["geographies", "locations"]

    batchfields = {
        "locations": [
            "id",
            "address",
            "match",
            "matchtype",
            "parsed",
            "coordinate",
            "tigerlineid",
            "side",
        ],
        "geographies": [
            "id",
            "address",
            "match",
            "matchtype",
            "parsed",
            "coordinate",
            "tigerlineid",
            "side",
            "statefp",
            "countyfp",
            "tract",
            "block",
        ],
    }

    def __init__(self, benchmark=None, vintage=None):
        """
        Arguments:
            benchmark (str): A name that references the version of the locator to use.
                See https://geocoding.geo.census.gov/geocoder/benchmarks
            vintage (str): The geography part of the desired vintage.
                See: https://geocoding.geo.census.gov/geocoder/vintages?form

        >>> CensusGeocode(benchmark='Public_AR_Current', vintage='Current_Current')
        """
        self._benchmark = benchmark or DEFAULT_BENCHMARK
        self._vintage = vintage or DEFAULT_VINTAGE

    def _geturl(self, searchtype, returntype=None):
        """Construct an URL for the geocoder."""
        returntype = returntype or self.returntypes[0]
        return self._url.format(returntype=returntype, searchtype=searchtype)

    def _fetch(self, searchtype, fields, **kwargs):
        """Fetch a response from the Geocoding API."""
        fields["vintage"] = self.vintage
        fields["benchmark"] = self.benchmark

        fields["format"] = "json"

        if "layers" in kwargs:
            fields["layers"] = kwargs["layers"]

        returntype = kwargs.get("returntype", "geographies")
        url = self._geturl(searchtype, returntype)

        try:
            with requests.get(url, params=fields, timeout=kwargs.get("timeout")) as r:
                content = r.json()
                if "addressMatches" in content.get("result", {}):
                    return AddressResult(content)

                if "geographies" in content.get("result", {}):
                    return GeographyResult(content)

                raise ValueError()

        except (ValueError, KeyError):
            raise ValueError("Unable to parse response from Census")

        except RequestException as err:
            raise err

    def coordinates(self, x, y, **kwargs):
        """Geocode a (lon, lat) coordinate."""
        kwargs["returntype"] = "geographies"
        fields = {"x": x, "y": y}

        return self._fetch("coordinates", fields, **kwargs)

    def address(self, street, city=None, state=None, zipcode=None, **kwargs):
        """Geocode an address."""
        fields = {
            "street": street,
            "city": city,
            "state": state,
            "zip": zipcode,
        }

        return self._fetch("address", fields, **kwargs)

    def onelineaddress(self, address, **kwargs):
        """Geocode an an address passed as one string.
        e.g. "4600 Silver Hill Rd, Suitland, MD 20746"
        """
        fields = {
            "address": address,
        }

        return self._fetch("onelineaddress", fields, **kwargs)

    def set_benchmark(self, benchmark):
        """Set the Census Geocoding API benchmark the class will use.
        See https://geocoding.geo.census.gov/geocoder/vintages?form for more."""
        self._benchmark = benchmark

    @property
    def benchmark(self):
        """Give the Census Geocoding API benchmark the class is using.
        See https://geocoding.geo.census.gov/geocoder/benchmarks for more."""
        return getattr(self, "_benchmark")

    def set_vintage(self, vintage):
        """Set the Census Geocoding API vintage the class will use.
        See https://geocoding.geo.census.gov/geocoder/vintages?form for more."""
        self._vintage = vintage

    @property
    def vintage(self):
        """Give the Census Geocoding API vintage the class is using.
        See https://geocoding.geo.census.gov/geocoder/vintages?form for more."""
        return getattr(self, "_vintage")

    def _parse_batch_result(self, data, returntype):
        """Parse the batch address results returned from the Census Geocoding API"""
        try:
            fieldnames = self.batchfields[returntype]
        except KeyError as err:
            raise ValueError("unknown returntype: {}".format(returntype)) from err

        def parse(row):
            row["lat"], row["lon"] = None, None

            if row["coordinate"]:
                try:
                    row["lon"], row["lat"] = tuple(float(a) for a in row["coordinate"].split(","))
                except:
                    pass

            del row["coordinate"]
            row["match"] = row["match"] == "Match"
            return row

        # return as list of dicts
        with io.StringIO(data) as f:
            reader = csv.DictReader(f, fieldnames=fieldnames)
            return [parse(row) for row in reader]

    def _post_batch(self, data=None, f=None, **kwargs):
        """Send batch address file to the Census Geocoding API"""
        returntype = kwargs.get("returntype", "geographies")
        url = self._geturl("addressbatch", returntype)

        if data is not None:
            # For Python 3, compile data into a StringIO
            f = io.StringIO()
            writer = csv.DictWriter(f, fieldnames=["id", "street", "city", "state", "zip"])
            for i, row in enumerate(data):
                row.setdefault("id", i)
                writer.writerow(row)

            f.seek(0)

        elif f is None:
            raise ValueError("Need either data or a file for CensusGeocode.addressbatch")

        try:
            form = MultipartEncoder(
                fields={
                    "vintage": self.vintage,
                    "benchmark": self.benchmark,
                    "addressFile": ("batch.csv", f, "text/plain"),
                }
            )
            headers = {"Content-Type": form.content_type}

            with requests.post(url, data=form, timeout=kwargs.get("timeout"), headers=headers) as r:
                # return as list of dicts
                return self._parse_batch_result(r.text, returntype)

        except RequestException as err:
            raise err

        finally:
            f.close()

    def addressbatch(self, data, **kwargs):
        """
        Send either a CSV file or data to the addressbatch API.
        According to the Census, "there is currently an upper limit of 1000 records per batch file."
        If a file, must have no header and fields id,street,city,state,zip
        If data, should be a list of dicts with the above fields (although ID is optional)
        """
        # Does data quack like a file handle?
        if hasattr(data, "read"):
            return self._post_batch(f=data, **kwargs)

        # If it is a string, assume it's a filename
        if isinstance(data, str):
            with open(data, "rb") as f:
                return self._post_batch(f=f, **kwargs)

        # Otherwise, assume a list of dicts
        return self._post_batch(data=data, **kwargs)


class GeographyResult(dict):

    """Wrapper for geography objects returned by the Census Geocoding API"""

    _coordkeys = ("CENTLON", "CENTLAT", "INTPTLON", "INTPTLAT")

    def __init__(self, data):
        self.input = data["result"].get("input", {})
        super().__init__(data["result"]["geographies"])

        # create float coordinate tuples
        for geolist in self.values():
            for geo in geolist:
                try:
                    geo["CENT"] = float(geo["CENTLON"]), float(geo["CENTLAT"])
                except ValueError:
                    geo["CENT"] = ()

                try:
                    geo["INTPT"] = float(geo["INTPTLON"]), float(geo["INTPTLAT"])
                except ValueError:
                    geo["INTPT"] = ()


class AddressResult(list):

    """Wrapper for address objects returned by the Census Geocoding API"""

    def __init__(self, data):
        self.input = data["result"].get("input", {})
        super().__init__(data["result"]["addressMatches"])



#!/usr/bin/env python3
# Copyright (C) 2015-7 Neil Freeman

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Command-line interface for censusgeocode"""
from __future__ import print_function

import argparse
import csv
import io
import sys

from . import __version__
from .censusgeocode import DEFAULT_BENCHMARK, DEFAULT_VINTAGE, CensusGeocode


def main():
    """Command-line interface for censusgeocode"""
    parser = argparse.ArgumentParser("censusgeocode", description="Command-line interface for the Census Geocoding API")

    parser.add_argument("-v", "--version", action="version", version="%(prog)s v" + __version__)
    parser.add_argument("address", type=str, nargs="?", default=None)
    parser.add_argument(
        "--csv",
        type=str,
        help=(
            "comma-delimited file of addresses. No header. Must have the following columns: id, street address, city, state, zip. "
            "The id must be a unique. Read from stdin with -"
        ),
    )
    parser.add_argument(
        "--rettype",
        choices=["locations", "geographies"],
        default="locations",
        help=(
            "Query type. Geographies will return state, county, tract, and block code in addition to TIGER/Line info and "
            "latitude and longitude. For use with --csv"
        ),
    )
    parser.add_argument(
        "--benchmark",
        default=DEFAULT_BENCHMARK,
        help="Version of the locator to query. See Census documentation for options.",
    )
    parser.add_argument(
        "--vintage",
        default=DEFAULT_VINTAGE,
        help="Geography version query. See Census documentation for options.",
    )
    parser.add_argument(
        "--timeout",
        metavar="SECONDS",
        type=int,
        default=12,
        help="Request timeout [default: 12]",
    )

    args = parser.parse_args()
    cg = CensusGeocode(benchmark=args.benchmark, vintage=args.vintage)

    if args.address:
        result = cg.onelineaddress(args.address, returntype=args.rettype, timeout=args.timeout)

        try:
            print("{},{}".format(result[0]["coordinates"]["x"], result[0]["coordinates"]["y"]))

        except IndexError:
            print("Address not found: {}".format(args.address), file=sys.stderr)

    elif args.csv:
        if args.csv == "-":
            # No streaming here - consume the entirety of stdin.
            infile = io.StringIO()
            csv.writer(infile).writerows(csv.reader(sys.stdin))
            infile.seek(0)

        else:
            infile = args.csv

        result = cg.addressbatch(infile, returntype=args.rettype, timeout=args.timeout)

        fieldnames = cg.batchfields[args.rettype] + ["lat", "lon"]
        fieldnames.pop(fieldnames.index("coordinate"))
        writer = csv.DictWriter(sys.stdout, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(result)

    else:
        print("Address or csv file required")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of censusgeocode.
# https://github.com/fitnr/censusgeocode

# Licensed under the General Public License (version 3)
# http://opensource.org/licenses/LGPL-3.0
# Copyright (c) 2015-9, Neil Freeman <contact@fakeisthenewreal.org>

from .censusgeocode import CensusGeocode

__version__ = '0.5.1'

cg = CensusGeocode()

coordinates = cg.coordinates
address = cg.address
onelineaddress = cg.onelineaddress
addressbatch = cg.addressbatch



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of censusgeocode.
# https://github.com/fitnr/censusgeocode

# Licensed under the General Public License (version 3)
# http://opensource.org/licenses/LGPL-3.0
# Copyright (c) 2015-9, Neil Freeman <contact@fakeisthenewreal.org>

from setuptools import setup

with open('censusgeocode/__init__.py') as i:
    version = next(r for r in i.readlines() if '__version__' in r).split('=')[1].strip('"\' \n')

with open('README.md') as f:
    readme = f.read()

setup(
    name='censusgeocode',
    version=version,
    description='Thin Python wrapper for the US Census Geocoder',
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords='census geocode api',
    author='Neil Freeman',
    author_email='contact@fakeisthenewreal.org',
    url='https://github.com/fitnr/censusgeocode',
    license='GNU General Public License v3 (GPLv3)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'requests[security]>=2.18,<3',
        'requests-toolbelt >=0.8.0, <1',
        'six'
    ],
    tests_require=[
        'vcrpy>=1.11.1, <1.12',
    ],
    packages=['censusgeocode'],
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'censusgeocode=censusgeocode.__main__:main',
        ],
    },
)
